/*=================================================================
 * 8/5/10 REA modified to remove round() and floor() funcs since got link error
 8/16/10 this version (zBackproject.c) handles complex data
 * Backproject.c
 * 
 * This .mex file performs the linear and nearest-neighbour
 * interpolation versions of the backprojection in iradon.m.
 * It does the same operation as the following piece of
 * code taken from iradon.m
 *
 * For nearest-neighbour interpolation
 *         for i=1:length(theta)   
 *            proj = p_real(:,i);
 *            t = round(x*costheta(i) + y*sintheta(i));
 *            img = img + proj(t+ctrIdx);
 *         end
 *
 * For linear interpolation
 *         for i=1:length(theta)
 *             proj = p_real(:,i);
 *             t = x.*costheta(i) + y.*sintheta(i);
 *             a = (t);
 *             img = img + (t-a).*proj(a+1+ctrIdx) + (a+1-t).*proj(a+ctrIdx);
 *         end
 *
 * Each loop can be replaced with one of
 *
 *         img = Backproject(p_real, costheta, sintheta, N, 0); % NN-interp
 *         img = Backproject(p_real, costheta, sintheta, N, 1); % linear interp
 *
 * The input parameters are:
 *
 *         p_real          Radon transform, with projections stored in columns
 *                    (as output by radon.m, for example)
 *         costheta   a vector (column or row) holding the cosine of the
 *                    projections angles
 *         sintheta   a vector (column or row) holding the sine of the
 *                    projections angles
 *         N          the output image with be NxN (this is a double float)
 *         interp     0 => nearest-neighbour interpolation
 *                    1 => linear interpolation
 *
 * The output is:
 *
 *         img        an NxN (double float) matrix
 *
 * WARNING: This code has not been exhaustively tested.  Use at your
 *          own risk. (standard disclaimer)
 *
 * Written by:
 *   Jeff Orchard
 *   Assistant Professor, David R. Cheriton School of Computer Science
 *   University of Waterloo
 *   Waterloo, Ontario, CANADA
 *   e-mail: jorchard@cs.uwaterloo.ca
 *   Web: http://www.cs.uwaterloo.ca/~jorchard/
 *
 * November 1, 2006
 *
 * This is a MEX-file for MATLAB.  
 * Copyright 1984-2004 The MathWorks, Inc.
 *
 *=================================================================*/
/* $Revision: 1.0 $ */
#include <math.h>
#include "mex.h"



void mexFunction( int nlhs, mxArray *plhs[], 
		  int nrhs, const mxArray*prhs[] )
     
{
	/* INPUT PARAMETERS */
        double *p_real,*p_imag;			/* filtered projections (one per col) */
	double *costheta;	/* pre-computed cos of proj_real angles */
	double *sintheta;	/* pre-computed sin of proj_real angles */
	double *Nptr;		/* size of reconstructed image */
	double *interp_ptr;	/* 0 for NN interp, 1 for linear interp */
	
	/* OUTPUT PARAMETERS */
	double *img_real,*img_imag;		/* output image */
	
	
	/* Other variables */
	int N;				/* integer copy of Nptr (above) */
	int interp_flag;	/* integer copy of interp_ptr (above) */
	int x_idx, y_idx, theta_idx;	/* loop indecies */
	double x;			/* stores x-coordinate of pixel */
	/*double y;*/		/* stores y-coordinate of pixel (not needed) */
	double ctr, xleft, ytop;	/* used to tranform from matrix indecies */
								/* to (x,y) coords (see iradon.m) */
	int len;		/* length of each projection (spatial dimension) */
	int n_proj;		/* number of projection angles (length of costheta vector, eg) */
	int ctr_idx;	/* centre index for projections */
	int i,j; /* for counters, offsets, etc. */
	
	/* Temporary variable used for code optimization */
	double cos_theta, sin_theta, t;
	int ctr_idx_1;	/* = ctr_idx + 1 (for optimization purposes) */
	double x_cos_theta;	/* = x * cos_theta */
	int ia;
	double *proj_real,*proj_imag;	/* points at the start of ia projection (ia column) */
	double *img_ptr_real,*img_ptr_imag;
	
    
    /* Check for proper number of arguments */
    
    if (nrhs != 5) { 
		mexErrMsgTxt("Five input arguments required."); 
    } else if (nlhs > 1) {
		mexErrMsgTxt("Too many output arguments."); 
    } 
	
    p_real = mxGetPr(prhs[0]);
    p_imag = mxGetPi(prhs[0]);

    costheta = mxGetPr(prhs[1]);
    sintheta = mxGetPr(prhs[2]);
    Nptr = mxGetPr(prhs[3]);
	interp_ptr = mxGetPr(prhs[4]);
	
	N = (int)(Nptr[0]);
	interp_flag = (int)(interp_ptr[0]);
	len = mxGetM(prhs[0]);
	n_proj = mxGetM(prhs[1]) * mxGetN(prhs[1]); /* use product to account for row- or col-vectors */
	
	/* The line below takes care of the difference between Matlab matrix indexing
	 * (base-1) and C-style matrix indexing (base-0).  The corresponding Matlab code is
	 *       ctr = ((N + 1)/2);
	 * eg. if N=8
	 *          Matlab -> ctr = index of 4
	 *          C      -> ctr = index of 3
	 */
	ctr = ((N-1) / 2);
	
	xleft = -ctr;
	ytop = ctr;
	
	ctr_idx = (int)(len/2);  /* centre index for projections */
	ctr_idx_1 = ctr_idx + 1;
    
    
    /* Create a matrix for the return argument */
    plhs[0] = mxCreateDoubleMatrix(N, N, mxCOMPLEX);
    img_real = mxGetPr(plhs[0]);  /* Get pointer to the data array */
    img_imag = mxGetPi(plhs[0]);  /* Get pointer to the data array */
	
	
    
    for (theta_idx=0 ; theta_idx<n_proj ; theta_idx++)
	{
	
		cos_theta = costheta[theta_idx];
		sin_theta = sintheta[theta_idx];
		i = theta_idx*len; // offset into proper column
		proj_real = (p_real+i);  /* point at proper column */
		proj_imag = (p_imag+i);  /* point at proper column */
		img_ptr_real = img_real;
		img_ptr_imag = img_imag;
		
		/* The loops below are designed to fly through the elements of the output image (img_real) in
		 * linear-memory sequence... makes it faster.  The matricies are stored column-by-column.
		 * The switch statement is outside the loops for optimization purposes.
		 */		
		
		switch (interp_flag)
		{
		
		case 0: /* nearest-neighbour interpolation */
		
			for (x_idx=0 ; x_idx<N ; x_idx++)
			{
				x = xleft + x_idx;  /* x-coord */
				x_cos_theta = x*cos_theta;  /* perform this multiplication outside the y-loop */
				t = x_cos_theta + ytop*sin_theta;  /* After this, t can simply be decremented by sin_theta each y-iter */
			
				for (y_idx=0 ; y_idx<N ; y_idx++)
				{
					/*y = ytop - y_idx; */
					/*t = x_cos_theta + y*sin_theta;*/  /* removed for optimization purposes */
				
					ia = floor(t + 0.5); //round
				        j = ia+ctr_idx;
					*img_ptr_real++ += proj_real[ia];
					*img_ptr_imag++ += proj_imag[ia];
				
					t -= sin_theta;  /* decrement t */
				} /* end of y-loop */
			} /* end of x-loop */
			
			break;
			
		case 1: /* linear interpolation */
		
			for (x_idx=0 ; x_idx<N ; x_idx++)
			{
				x = xleft + x_idx;  /* x-coord */
				x_cos_theta = x*cos_theta;  /* perform this multiplication outside the y-loop */
				t = x_cos_theta + ytop*sin_theta;  /* After this, t can simply be decremented by sin_theta each y-iter */
			
				for (y_idx=0 ; y_idx<N ; y_idx++)
				{
					/*y = ytop - y_idx;
					/*t = x_cos_theta + y*sin_theta;*/  /* removed for optimization purposes */
				
					ia = floor(t); 
				
					/* The following line is the slower version (2 multiplications) */
					/**img_ptr_real++ += (t-ia)*proj_real[ia+ctr_idx_1] + (1-t+ia)*proj_real[ia+ctr_idx]; */
				        i = ia+ctr_idx_1;
				        j = ia+ctr_idx;
					/* This is the speedy version (1 multiplication) */
					*img_ptr_real++ += (t-ia)*( proj_real[i] - proj_real[j] ) + proj_real[j];
					*img_ptr_imag++ += (t-ia)*( proj_imag[i] - proj_imag[j] ) + proj_imag[j];
				
					t -= sin_theta;  /* decrement t */
				} /* end of y-loop */
			} /* end of x-loop */
			
			break;
			
		} /* end of switch statement */
	} /* end of theta loop */
	
  	
    return;
    
}


